/*
 * MySocket.cpp: Example Socket Server.
 *
 * Copyright (C) 2007 - Almer S. Tigelaar
 *
 * (based mostly on some Linux related socket API books & some
 *  only resources for the Windows (WinSock) 'flavour').
 *
 * Usage: Run it and telnet to local port 9999. Whatever you type
 * in your telnet session will be echoed back.
 * Note: The program only accepts (and handles) one client connection.
 */

/* Franois: You can find the includes that were used in the stdafx.h
 * For this networking stuff you only need winsock.h.
 */
#include "TPPSocket.h"
#include "LineSocketClient.h"
#include "LineSocketServer.h"

static LineSocketServer* socketHandler = NULL;

static void CALLBACK timerProc(HWND hwnd, UINT message, UINT idTimer, DWORD dwTime)
{ 
	std::cout << "* Tick" << std::endl;
	
	if (!socketHandler->hasClient()) { /* No connection (yet) */
		return;
	}

	if (socketHandler->hasLine()) {
		/* Read line batch & send done */
		do {
			std::string line = socketHandler->getLine();
			std::cout << "Read: " << line << std::endl;
		} while (socketHandler->hasLine());

		socketHandler->sendLine("DONE\n");
		/* In fact this emulation is not entirely accurate, since we should
	     * send done for each [TTS] block encountered. So, we risk sending to little or
	     * too few "DONE"'s, nevertheless this doesn't seem to be much of a problem when
	     * running on the same system and it works fine for debugging purposes.
	     */
	}
}

/*int _tmain(int argc, _TCHAR* argv[])
{
	LineSocket* lineSocket = new LineSocket(9999);

	while (true) {
		lineSocket->waitForClient();
		while (lineSocket->hasClient()) {
			if (lineSocket->hasLine()) {
				std::string line = lineSocket->getLine();
				std::cout << "Read: " << line << std::endl;
			}

			Sleep(1000);
		}
	}
}*/

static void server()
{
	socketHandler = new LineSocketServer(9999);
	socketHandler->setNoDelay(true);

	int timerID = SetTimer(NULL, NULL, 2000, (TIMERPROC) timerProc);
	std::cout << "Timer ID: " << timerID << std::endl;

	while (true) {
		if (!socketHandler->hasClient()) {
			socketHandler->waitForClient();
		}
		
		MSG msg;
		while(GetMessage(&msg, NULL, 0, 0))
		{
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}
	delete socketHandler; /* Unreachable, but cute */

}

static void client()
{
	LineSocketClient* clientSocketHandler = new LineSocketClient("127.0.0.1", 9999);
	clientSocketHandler->setNoDelay(true);

	while (true) {
		std::cout << "Sending Line ..." << std::endl;
		clientSocketHandler->sendLine("Test\n");
		Sleep(1000);
		while (clientSocketHandler->hasLine()) {
			std::cout << clientSocketHandler->getLine() << std::endl;
		}
	}

	delete clientSocketHandler;
}

int main(int argc, char* argv[])
{
	server();
	return 0;
}

