// VisageLinkDoc.cpp : implementation of the CVisageLinkDoc class
//

#include "stdafx.h"
#include "VisageLink.h"

#include "VisageLinkDoc.h"

/*
#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif
*/

#include "OpenGLWnd.h"
#include "math.h"
#include "MainFrm.h"
#include "SimpleFacialExpression.h"
#include "SimpleFbaAction.h"


/////////////////////////////////////////////////////////////////////////////
// CVisageLinkDoc

IMPLEMENT_DYNCREATE(CVisageLinkDoc, CDocument)

BEGIN_MESSAGE_MAP(CVisageLinkDoc, CDocument)
	//{{AFX_MSG_MAP(CVisageLinkDoc)
	ON_COMMAND(ID_COMMAND_REINIT_SOCKET, reinitSocket)
	ON_COMMAND(ID_COMMAND_REINIT_AGENT, reinitAgent)
	ON_COMMAND(ID_FILE_START, OnFileStart)
	ON_COMMAND(ID_FILE_SPEAK, OnFileSpeak)
	ON_COMMAND(ID_OPTIONS_GAZEFOLLOWINGVOR, OnOptionsGazefollowingvor)
	ON_COMMAND(ID_OPTIONS_DISABLEGAZEFOLLOWINGVOR, OnOptionsDisablegazefollowingvor)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CVisageLinkDoc construction/destruction
CVisageLinkDoc* CVisageLinkDoc::instance = NULL; // singleton instance

CVisageLinkDoc::CVisageLinkDoc()
{
	animationStarted = false;
	m_FAPlayer = new FAPlayer();
	vor = new SimpleVORAction();
	instance = this; // Set singleton instance

	char * sectionname = "Settings";

	CWinApp * app = AfxGetApp();

	port = app->GetProfileInt("Settings", "Port", 9999);
	model = app->GetProfileString(sectionname, "Model", "kenneth.afm");
	voice = app->GetProfileString(sectionname, "Voice", "ScanSoft Tom_Full_22kHz");

	this->socketHandler = NULL;
}

CVisageLinkDoc::~CVisageLinkDoc()
{
}

BOOL CVisageLinkDoc::OnNewDocument()
{
	if (!CDocument::OnNewDocument())
		return FALSE;

	// TODO: add reinitialization code here
	// (SDI documents will reuse this document)

	return TRUE;
}

BOOL CVisageLinkDoc::OnOpenDocument(LPCTSTR lpszPathName) 
{
	return TRUE;
}



/////////////////////////////////////////////////////////////////////////////
// CVisageLinkDoc serialization

void CVisageLinkDoc::Serialize(CArchive& ar)
{
	if (ar.IsStoring())
	{
		// TODO: add storing code here
	}
	else
	{
		// TODO: add loading code here
	}
}



/**
* A simple wave action.
* This is an implementation of an extremely simple hand waving action, actually just a very basic movement of the
* right arm.
*/
class SimpleWaveAction : public SimpleFbaAction
{
public:
	FBAPs *getFBAPs(long globalTime, FBAPs *lastFBAPs, AFM *afm)
	{
		long localTime = globalTime - globalStartTime;
		float angle = 0.0f;
		
		
		if(!fbaps)
			fbaps = new FBAPs();
		
		
		angle = -sin((1000+localTime)/300.0f) * 300000;
		
		fbaps->setBAP(r_shoulder_flexion,(int)angle);
		
		if(localTime < 500)
			return(fbaps);
		else
			return NULL;
	};
};


/**
* Processing of the TTS viseme notification events.
* This function is the implementation of a virtual function from the Visagesapi5ttsObserver class.
* It is not to be called directly. Rather, the instance of CVisageLinkDoc is attached to the
* Visagesapi5tts, which calls the Notify() function as appropriate, and passes the appropriate
* parameters to it.
*
* This gives the possibility of additional control over the animation parameters generated by TTS.
* In this example we do not use it so it is empty.
*
* @param faps pointer to FAPs structure, or NULL at the end of speech.
* @param cp pointer to the coding parameters used when encoding the animation into a file; 
* @param pts current TTS speech bit stream time in milliseconds. 
* @param vtts instance of Visagesapi5tts that called the function; this can be used for identification if more than one Visagesapi5tts object is used at the same time
*
* @see Visagesapi5ttsObserver
* @see preRender()
*/
void  CVisageLinkDoc::Notify (Visagesapi5tts *vtts, FBAPs *fbaps, CodingParameters *cp, long pts)
{
}

/**
* Processing of the TTS bookmark events.
* This function is the implementation of a virtual function from the Visagesapi5ttsObserver class.
* It is not to be called directly. Rather, the instance of CVisageLinkDoc is attached to the
* Visagesapi5tts, which calls the BookMark() function as appropriate, and passes the appropriate
* parameters to it.
*
* In this example we use the bookmarks to insert simple facial expressions or animations at particular times
* during the speech. This is done by playing a SimpleFacialExpression. The expression is
* initialized with the name from the bookmark tag, e.g. "surprise" or "joy". See SimpleFacialExpression
* for the list of supported expression names.
*
* The other bookmark that can occur is "wave", and if it is found we insert a simple waving animation - just moving the right hand.
*
* @param qTimeStamp bookmark time stamp in milliseconds. This is bookmark's TTS speech bit stream time.
* @param dwMarkNum bookmark code number (the number that was entered in the SAPI control tag as \\Mrk=number\.
* @param vtts instance of Visagesapi5tts that called the function; this can be used for identification if more than one Visagesapi5tts object is used at the same time
*
* @see Visagesapi5ttsObserver
* @see SimpleFacialExpression
*/
void  CVisageLinkDoc::BookMark (Visagesapi5tts *vtts, long qTimeStamp, DWORD dwMarkNum, WCHAR *markText)
{
	CString t(markText);
	if(!t.CompareNoCase("wave"))
	{
		SimpleWaveAction *wave = new SimpleWaveAction();
		m_FAPlayer->playTrack(wave);
	}
	else
	{
		SimpleFacialExpression *exp = new SimpleFacialExpression((char*)markText,2200,1.3f);
		m_FAPlayer->playTrack(exp);
	}
}

/**
* Processing of the TTS start speech events.
* This function is the implementation of a virtual function from the Visagesapi5ttsObserver class.
* It is not to be called directly. Rather, the instance of CVisageLinkDoc is attached to the
* Visagesapi5tts, which calls the StartSpeech() function as appropriate, and passes the appropriate
* parameters to it.
*
* In this example we use the bookmarks to insert simple expressions and animations at particular times
* during the speech. Some bookmarks are already inserted in the input text file. We use this function
* as a simple filter to insert additional bookmarks. 
* 
* So, when the string "surpr" is encountered, a bookmark for the surprize expression is inserted. 
*
* This is a very simple example of using the Visagesapi5ttsObserver for processing the text and 
* inserting appropriate actions.
*
* @param qTimeStamp bookmark time stamp in milliseconds. This is bookmark's TTS speech bit stream time.
* @param dwMarkNum bookmark code number (the number that was entered in the SAPI control tag as \\Mrk=number\.
* @param vtts instance of Visagesapi5tts that called the function; this can be used for identification if more than one Visagesapi5tts object is used at the same time
*
* @see Visagesapi5ttsObserver
* @see SimpleFacialExpression
*/
void CVisageLinkDoc::StartSpeech (Visagesapi5tts *vtts, WCHAR *text)
{
	CString t(text);
	int ind = 0;
	while(ind != -1)
	{
		ind = t.Find("surpr",ind);
		if(ind != -1)
		{
			t.Insert(ind,"<BOOKMARK MARK='surprise'/>");
			ind += 32;
		}
	}


	ind = t.GetLength();

	char *text1 = t.GetBuffer(ind);

	/* NOTE: I commented out this line since VS 2005 doesn't compile
	 * cleanly otherwise. It searches for a toUnicode definition with
	 * an other signature than the one in the visage5sapitts.lib file
	 * (which is "?toUnicode@Visagesapi5tts@@SAXPAGPAD@Z")
	 */
	//Visagesapi5tts::toUnicode(text,text1);
}

/////////////////////////////////////////////////////////////////////////////
// CVisageLinkDoc diagnostics

#ifdef _DEBUG
void CVisageLinkDoc::AssertValid() const
{
	CDocument::AssertValid();
}

void CVisageLinkDoc::Dump(CDumpContext& dc) const
{
	CDocument::Dump(dc);
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CVisageLinkDoc commands

/**
* Check if all necessary files are present.
* 
* All the file names in this example are hard-coded. This function checks if the files are actually
* present and gives error messages if they are not. The example should be run in the VisageLink folder where all necessary files are present.
*/
void CVisageLinkDoc::CheckFiles() 
{
		FILE * testFP;

		if(!(testFP = fopen("p2v.cfg","r")))
		{
			CString msg;
			msg = "The working folder does not contain the file p2v.cfg, necessary for the TTS.";
			MessageBox(0,msg,"Error",MB_ICONERROR);
			exit(1);
		}
		else fclose(testFP);

	if(!(testFP = fopen(model,"r")))
	{
		CString msg;
		msg = "The working folder does not contain the file "+model+", necessary to run this example.";
		MessageBox(0,msg,"Error",MB_ICONERROR);
		exit(1);
	}
	else fclose(testFP);
	
}

/*
 * NOTE (Almer): This is a highly inelegant timer based solution. Nevertheless it does
 * the job. Basically it just checks if anything is waiting on the socket every timer interval
 * and if so, it pushes it to processLine that handles the actual 'talking'.
 * A semi-singleton construction was used here (the instance variable) to work around
 * a win32 API limitation (which does not want the timer callback to be part of a class).
 *
 * Better solutions:
 * We can use the winsock2 WSAEvent callback, I've experimented with this, but could not
 * get it to work easily (without converting the entire TPPSocket to the Winsock2 api). 
 * Using threads is not really a good solution I think, because it makes our application overly
 * complicated (it is not really necessary if we can use the Event mechanisms in Winsock 2).
 *
 * Anyway, we can look at that at a later time, the solution as it stands introduces no more than
 * the timer interval of additional delay (which we can set really low).
 */
static void CALLBACK timerProc(HWND hwnd, UINT message, UINT idTimer, DWORD dwTime)
{ 
	if (!CVisageLinkDoc::instance->socketHandler->hasClient()) { // No client is connected
		return;
	}
	
	// Read all pending lines from client and process them.
	while (CVisageLinkDoc::instance->socketHandler->hasLine()) {
		std::string line = CVisageLinkDoc::instance->socketHandler->getLine();

		/*string msg = "Read From Socket: ";
		msg.append(line);
		MessageBox(0, msg.c_str(), "Info", MB_ICONINFORMATION);*/

		CVisageLinkDoc::instance->processLine(line);
	}
}

/**
* Initialize everything and start the animation.
* 
* This method is called when the File->Start menu option is selected. It performs all necessary initializations, and starts the FAPlayer.
*
* Specifically:
* - a check is performed to see if all required files are present
* - a face/body model is loaded and attached to the FAPlayer
* - a base animation track (simple head motion and eye blinks) is loaded from file into the player; it will loop forever. This makes the animation look nicer, and provides some basic action to the FAPlayer so that it continues playing (if the player is empty, the play() method would just return because there is nothing to play)
* - the speech synthesis object vtts is initialised and added as an animation track to the player
* - the rendering mechanism is started
* - the FAPlayer play() method is called to start the player.
*/
void CVisageLinkDoc::reinitAgent() {
	animationStarted = true;
	
	CheckFiles();         // check if all necessary files are present

	//=-- Get renderer
	CMainFrame* pMainWnd = (CMainFrame*)AfxGetMainWnd();
	COpenGLWnd *pOpenGL = pMainWnd->GetOpenGLWnd();


	if (m_FAPlayer != NULL) {
		//=-- Remove previous model
//		pOpenGL->Close();
		pOpenGL->RemoveModel(m_FAPlayer->getFaceModel());
		m_FAPlayer->~FAPlayer();
	}
	m_FAPlayer = new FAPlayer();
	AFM *afm = new AFM(model.GetBuffer(model.GetLength()));          // load the face/body model
	m_FAPlayer->init(afm);              // add the model to the player, and add this CVisageLinkDoc as renderer
//	m_FAPlayer->setSleepTimePerCycle(100);

	FbaFileAction *trk = m_FAPlayer->addTrack("baseheadmotion.fba", 1, 0);    // start playing the base animation
	trk->smoothZeroValues(); // smooth the zero values to avoid jerks when merging with other actions

	vtts = new Visagesapi5tts(); // construct a TTS
	vtts->setSpeaker(voice.GetBuffer(voice.GetLength())); 

	m_FAPlayer->addTrack(vtts);// add vtts as an animation track to the FAPlayer
	vtts->attach(this);             // attach this CVisageLinkDoc to the vtts as observer; 
	                                // it will get viseme and bookmark events; 
	                                // bookmark events are used to insert expressions
	
	// initialise the renderer with the new face model
	pOpenGL->Init(m_FAPlayer->getFaceModel()); 
	
	m_FAPlayer->play();      // start the player. 
	return;
}

void CVisageLinkDoc::OnFileStart() 
{
	 return;
}


void CVisageLinkDoc::processLine(std::string line) 
{
	// Extract Command and subline
	if (line.size()>4)
	{
	string command = line.substr(1,3);
	string subline = line.substr(5);

		// TTS Command given
		if (!command.compare("TTS"))
		{
			char* param = const_cast<char*>(subline.c_str());
			vtts->speak(param, 0);
			this->socketHandler->sendLine("DONE\n");
			return;
		}

		// FBA Command given
		else if (!command.compare("FBA"))
		{
			char* param = const_cast<char*>(subline.c_str());
			m_FAPlayer->playTrack(param);
			return;
		}

		// CTL Command given
		else if (!command.compare("CTL"))
		{ 
			// Extract Subcommand
			string subcommand = subline.substr(1,3);
			string subsubline = subline.substr(5);

				// SPD Subcommand given
				if (!subcommand.compare("SPD"))
				{
					int param;
					sscanf(subsubline.c_str(), "%d", &param);
					vtts->setSpeed(param);
					return;
				}

				// PIT Subcommand given
				else if (!subcommand.compare("PIT"))
				{
					int param;
					sscanf(subsubline.c_str(), "%d", &param);
					vtts->setPitch(param);
					return;
				}

				// VOL Subcommand given
				else if (!subcommand.compare("VOL"))
				{
					int param;
					sscanf(subsubline.c_str(), "%d", &param);
					
					// Normalize input
					int steps = 20;
					param = (65536 / steps) * param;
					param = (65536 * param) + param;

					waveOutSetVolume(0, (DWORD)param);
					return;
				}

		}	

		else if (!command.compare("CFG"))
		{ 
			// Extract Subcommand
			string subcommand = subline.substr(1,3);
			string subsubline = subline.substr(5);

			if		(!subcommand.compare("FAC")) { this->model	= subsubline.c_str(); }
			else if (!subcommand.compare("VOC")) { this->voice	= subsubline.c_str(); }
//			else if (!subcommand.compare("VCP")) { this->vcp	= subsubline.c_str(); }
//			else if (!subcommand.compare("ABP")) { this->abp	= subsubline.c_str(); }
			else if (!subcommand.compare("COM")) { this->reinitAgent(); }
			else if (!subcommand.compare("LST")) { this->socketHandler->sendLine("DONE\n"); }

			return;
		}	
	}

	// Fallback routine: just say it
	char* param = const_cast<char*>(line.c_str());
	vtts->speak(param, 0); 
	this->socketHandler->sendLine("DONE\n");

	return;
}



void CVisageLinkDoc::initSocket() {
	//=-- Open socket connection
	this->socketHandler = new LineSocketServer(port);
	this->socketHandler->setNoDelay(true); // Enable TCP_NODELAY, since we do not immediately send something back in our protocol
	if (!this->socketHandler->waitForClient()) { // Wait for a client to connect
		CString msg;
		msg = "Error while waiting for client!";
		MessageBox(0,msg,"Error",MB_ICONERROR);
		exit(1);
	}

	// Bind a timer, and set it to read the socket contents every 25 msecs.
	SetTimer(NULL, NULL, 25, (TIMERPROC) timerProc);

}

void CVisageLinkDoc::reinitSocket() {
	this->closeSocket();
	this->initSocket();
}

void CVisageLinkDoc::closeSocket() {
	if (this->socketHandler) {
		delete this->socketHandler;
		this->socketHandler = NULL; 
	}
}




/**
* Speak. OBSOLETE
*/
void CVisageLinkDoc::OnFileSpeak() 
{ return; }

/**
* Enable gaze following (vestibulo-ocular reflex).
* 
* This method is called when the File->Enable Gaze Following menu option is selected. 
*
* It enables the gaze following (vestibulo-ocular reflex) by attaching the vor object (a SimpleVORAction)
* to the FAPlayer.
*
* @see SimpleVORAction
*/
void CVisageLinkDoc::OnOptionsGazefollowingvor() 
{
	if(!vor) return;
	if(!m_FAPlayer) return;
	
	m_FAPlayer->addTrack(vor);
}

/**
* Disable gaze following (vestibulo-ocular reflex).
* 
* This method is called when the File->Disable Gaze Following menu option is selected. 
*
* It disables the gaze following (vestibulo-ocular reflex) by  removing the vor object (a SimpleVORAction)
* from the FAPlayer.
*
* @see SimpleVORAction
*/
void CVisageLinkDoc::OnOptionsDisablegazefollowingvor() 
{
	if(!vor) return;
	if(!m_FAPlayer) return;
	
	m_FAPlayer->removeTrack(vor);
}
