// VisageLinkDoc.h : interface of the CVisageLinkDoc class
//
/////////////////////////////////////////////////////////////////////////////

#if !defined(AFX_VisageLinkDOC_H__5C6AA0DA_63F2_4882_B589_7C56AC7E188B__INCLUDED_)
#define AFX_VisageLinkDOC_H__5C6AA0DA_63F2_4882_B589_7C56AC7E188B__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "visagesapi5tts.h"
#include "FAPlayer.h"
#include "AFM.h"
#include "SimpleVORAction.h"
#include "shlwapi.h"
#include "LineSocketServer.h"

/*! \mainpage  visage|SDK VisageLink
* VisageLink demonstrates the following capabilities of <a href="../../../../doc/index.html">visage|SDK</a>:
* - using a SAPI-5 Text-to-Speech engine to generate speech and corresponding face animation and 
* render them immediately on the screen 
* - using FAPlayer to dynamically insert additional actions (gestures, movements etc.) into the animation based on bookmarks in the input text using the Visagesapi5ttsObserver mechanism
*
* To use the application:
*
* - File->Start loads the virtual character and initializes the animation. Since all file names are hard-coded in this example, it is best to run
* the example in the VisageSDK/Samples/VisageLink folder, where all necessary files are present. If any files
* are missing, error will be reported. 
* - File->Speak  reads the text from the input file test-sapi5-expression.txt and uses TTS to generate
* speech and animation. Animation is rendered in real time on a face model immediately as it is generared, i.e. during the speech synthesis.
* Bookmarks in the text are interpreted as facial expressions. 
* The test-sapi5-expression.txt file includes a surprise expression towards the end
* of animation and a joy expression at the beginning, as well as a crude waving action.
* - Options menu items allow to enable and disable the gaze following, or vestibulo-ocular reflex. When it is enabled, the gaze
* is directed always straight ahead (like the character is watching you) even when the head rotates.
*
* In general, the code works as follows. A base animation track
* (headmotion_short.fba) is loaded and played in a loop by an FAPlayer. This track contains simple head motion.
* A Visagesapi5tts object is attached as an FbaAction to the FAPlayer. Therefore, when Visagesapi5tts::speak()
* method is called, the FAPlayer will play the animation generated by Visagesapi5tts.
* Visagesapi5tts also calls the Visagesapi5ttsObserver callback functions Notify() and BookMark(). These
* functions are implemented in CVisageLinkDoc class, which is an implementation of Visagesapi5ttsObserver. 
* In the BookMark() callback
* function we receive the notification of bookmarks inserted into the text. The BookMark() function is called 
* when the bookmark is encountered, and this mechanism allows us to insert animation actions at the
* desired time in the speech, e.g. just before a particular word is pronounced. This mechanism is used to 
* trigger basic expressions inserted into the text as bookmarks.
*
* All callbacks are implemented in the CVisageLinkDoc class, which
* implements the Visagesapi5ttsObserver abstract class.
*
* This example has initially been generated using the MS Visual C++ 6.0 AppWizard and most of
* it has been generated automatically. Please refer to the <a href="../ReadMe.txt"> ReadMe</a>
* file generated by the AppWizard for the general structure of the program. The code also
* contains plenty of comments generated by the AppWizard.
* 
* *****************IMPORTANT***********************************
*
* This project requires the installation of the Microsoft Speech SDK 5.1, 
* and setting the include and library paths accordingly.
* To download Microsoft Speech SDK 5.1 go to http://www.microsoft.com/downloads and search for Speech SDK, version 5.1; download SpeechSDK51.exe.
*
* *************************************************************
*
* The specific classes and methods that demonstrate visage|SDK are:
*
* - CVisageLinkDoc: Implements the necessary functions of Visagesapi5ttsObserver.
* - CVisageLinkDoc::OnFileStart(): Performs all initialization procedures.
* - CVisageLinkDoc::OnFileSpeak(): Reads the text from the file and speaks it.
* - CVisageLinkDoc::BookMark(): Processes bookmarks in the input text and inserts expressions based on the bookmarks.
* - SimpleFacialExpression: Implements a facial expression action with simple dynamics.
* - CVisageLinkDoc::OnOptionsGazefollowingvor: Enables the gaze following behavior.
*/

/**
* Main document class.
* Implements the necessary functions of FARenderer and Visagesapi5ttsObserver.
*/
class CVisageLinkDoc : public CDocument, public Visagesapi5ttsObserver
{
protected: // create from serialization only
	CVisageLinkDoc();
	DECLARE_DYNCREATE(CVisageLinkDoc)

// Attributes
public:
	/**
	* The FAPlayer.
	* The Face Animation Player (FAPlayer). It is used to load and play animation and corresponding sound.
	*/
	FAPlayer *m_FAPlayer;

	/**
	* The Visage TTS.
	* The TTS object used for speaking.
	*/
	Visagesapi5tts *vtts;

	/**
	* The Vestibolo-Ocular Reflex (gaze following) action.
	*/
	SimpleVORAction  *vor;

	/**
	* Animation started flag.
	* This flags indicates that the basic animation is started and everything is initialised and ready for speaking..
	*/
	bool animationStarted;

	LineSocketServer* socketHandler;
	static CVisageLinkDoc* instance;

// Operations
public:
	BOOL OnOpenDocument(LPCTSTR lpszPathName);

// Overrides
	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(CVisageLinkDoc)
	public:
	virtual BOOL OnNewDocument();
	virtual void Serialize(CArchive& ar);
	//}}AFX_VIRTUAL

// Implementation
public:
	void Notify (Visagesapi5tts *vtts, FBAPs *faps, CodingParameters *cp, long pts);
	void BookMark (Visagesapi5tts *vtts, long qTimeStamp, DWORD dwMarkNum,WCHAR *markText);
	void StartSpeech (Visagesapi5tts *vtts, WCHAR *text);
	void CVisageLinkDoc::CheckFiles();
	void processLine(std::string line);
	void initSocket();
	void reinitSocket();
	void closeSocket();
	void reinitAgent();

	virtual ~CVisageLinkDoc();
#ifdef _DEBUG
	virtual void AssertValid() const;
	virtual void Dump(CDumpContext& dc) const;
#endif

protected:

// Generated message map functions
protected:
	//{{AFX_MSG(CVisageLinkDoc)
	afx_msg void OnFileStart();
	afx_msg void OnFileSpeak();
	afx_msg void OnOptionsGazefollowingvor();
	afx_msg void OnOptionsDisablegazefollowingvor();
	//}}AFX_MSG
	DECLARE_MESSAGE_MAP()
private:
	int m_iLastRenderedTime;
	int port;
	CString model;
	CString voice;

};

/////////////////////////////////////////////////////////////////////////////

//{{AFX_INSERT_LOCATION}}
// Microsoft Visual C++ will insert additional declarations immediately before the previous line.

#endif // !defined(AFX_VisageLinkDOC_H__5C6AA0DA_63F2_4882_B589_7C56AC7E188B__INCLUDED_)
